import logging

from tengi.command.command_handler import *
from tengi.setup import config_utils


class CommandHandlerPassword(CommandHandler):
    def get_cards(self) -> Iterable[CommandCard]:
        return [CommandCard(command_str='/remember_password',
                            description='Admin. Remember admin password for current chat. --chat_id optional.',
                            is_admin=True),
                CommandCard(command_str='/forget_password',
                            description='Admin. Forget admin password for current chat. --chat_id optional.',
                            is_admin=True)]

    def handle(self, context: CommandContext):
        target_chat_id = context.sender_chat_id if (context.args.chat_id is None) else context.args.chat_id
        if context.command == '/remember_password':
            password = context.get_mandatory_arg('password')
            config_utils.remember_password(config=context.config,
                                           chat_id=target_chat_id,
                                           password=password)
            context.reply('Remembered', log_level=logging.INFO)
        elif context.command == '/forget_password':
            if config_utils.has_remembered_password(config=context.config, chat_id=target_chat_id):
                config_utils.delete_remembered_password(config=context.config, chat_id=target_chat_id)
                context.reply('Forgotten', log_level=logging.INFO)
            else:
                context.reply('Was not remembered')
        else:
            raise ValueError(f'Unhandled command: {context.command}')



import logging

from tengi.setup.app import App
from tengi.setup.config import Config
from tengi.telegram import telegram_bot_utils, telegram_api_utils
from tengi.telegram.telegram_bot import TelegramBot
from tengi.telegram.telegram_cursor import TelegramCursor
from tengi.telegram.inbox_hub import TelegramInboxHub
from tengi.telegram.inbox_handler import TelegramInboxHandler
from tengi.telegram import telegram_error
from tengi.telegram.telegram_api import TelegramApi
from tengi.telegram.reply_context import ReplyContext, ReplyContextTelegram, ReplyContextLogOnly
from tengi.command.command_handler import CommandHandler
from tengi.command.handler_essentials import CommandHandlerEssentials
from tengi.command.handler_password import CommandHandlerPassword
from tengi.command.handler_config import CommandHandlerConfig
from tengi.command.command_parser import CommandParser
from tengi.command.param import CommandParam
from tengi.command.card import CommandCard
from tengi.command.command_hub import CommandHub
from tengi.command import tengi_command_params
from tengi.command.handler_pool import CommandHandlerPool
from tengi.command.command_error import *
from tengi.hasher import Hasher
from tengi.csv_log.csv_logger import CsvLogger
from tengi.csv_log.messages_logger import MessagesLogger
from tengi.hack import jsonstore_hack
from tengi.state.preserver import Preserver
from tengi.state.timed_preserver import TimedPreserver
from tengi.state.abuse_detector import AbuseDetector
from tengi.state.abuse_janitor import AbuseJanitor
from tengi.state.chat_id_preserver import ChatIdPreserver

jsonstore_hack.fix_jsonstore_dumps()

logging.getLogger(__name__).addHandler(logging.NullHandler())




from typing import Iterable
from telebot import types as bot_types
from telethon.tl import types as api_types


def api_to_bot_markup(api_markup: api_types.ReplyInlineMarkup) -> bot_types.InlineKeyboardMarkup:
    bot_markup = bot_types.InlineKeyboardMarkup()
    if api_markup is not None:
        for api_r in api_markup.rows:
            bot_r = []
            for api_b in api_r.buttons:
                if isinstance(api_b, api_types.KeyboardButtonCallback):
                    bot_b = bot_types.InlineKeyboardButton(text=api_b.text,
                                                           callback_data=api_b.data.decode(encoding='utf-8'))
                elif isinstance(api_b, api_types.KeyboardButtonUrl):
                    bot_b = bot_types.InlineKeyboardButton(text=api_b.text,
                                                           url=api_b.url)
                else:
                    raise TypeError(f'Unhandled button type: {type(api_b)}')
                bot_r.append(bot_b)
            bot_markup.add(*bot_r)
    return bot_markup


def iterate_buttons(message: api_types.Message) -> Iterable[api_types.KeyboardButton]:
    if message.reply_markup is not None:
        for row in message.reply_markup.rows:
            for b in row.buttons:
                yield b



from telebot.types import Message
from typing import Iterable

from tengi.csv_log.csv_logger import *
from tengi import CommandParser, event, Hasher
from tengi.telegram.telegram_bot_utils import encode_case_id, get_file_id
from tengi.telegram.telegram_bot import EV_TEL_RECEIVED_MESSAGES, EV_TEL_SENT_MESSAGE


class MessagesLogger(CsvLogger):
    def __init__(self,
                 dir_path: Path,
                 file_name_prefix: str,
                 command_parser: CommandParser,
                 hasher: Hasher,
                 chat_types: Iterable[str] = None):
        super().__init__(dir_path=dir_path, file_name_prefix=file_name_prefix)

        self.command_parser = command_parser
        self.hasher = hasher
        self.chat_types = chat_types

        event.emitter.on(EV_TEL_RECEIVED_MESSAGES, self._on_receive_messages)
        event.emitter.on(EV_TEL_SENT_MESSAGE, self._on_self_send_message)

    def try_log_message(self, m: Message, is_self: bool, hide_command_password=True):
        if (self.chat_types is not None) and (m.chat.type not in self.chat_types):
            return

        case_id = encode_case_id(chat_id=m.chat.id,
                                 message_id=m.message_id)
        text = m.text
        if text is None:
            text = m.caption

        if hide_command_password:
            if self.command_parser.is_command(text) and self.command_parser.contains_secret(text):
                text = self.command_parser.hide_secret(text)

        row_dict = {
            'case_hash': self.hasher.trimmed(case_id),
            'conversation_hash': self.hasher.trimmed(m.chat.id),
            'received_dt': datetime.utcfromtimestamp(m.date),
            'is_self': is_self,
            'chat_type': m.chat.type,
            'content_type':  m.content_type,
            'text': text,
            'file_id': get_file_id(m),
        }

        self.write_row(row_dict)

    def _on_receive_messages(self, messages: Iterable[Message]):
        for m in messages:
            self.try_log_message(m, is_self=False, hide_command_password=True)

    def _on_self_send_message(self, m: Message):
        self.try_log_message(m, is_self=True, hide_command_password=False)



import logging
from telebot.apihelper import ApiTelegramException

from tengi.command.handler_pool import CommandHandlerPool
from tengi.command.command_parser import CommandParser
from tengi.setup.config import Config
from tengi.setup import config_utils
from tengi.telegram.telegram_bot import TelegramBot
from tengi.telegram.inbox_handler import *
from tengi.command.command_context import CommandContext
from tengi.command.command_error import CommandMissingArgError

logger = logging.getLogger(__file__)


class CommandHub(TelegramInboxHandler):
    def __init__(self,
                 config: Config,
                 telegram_bot: TelegramBot,
                 parser: CommandParser,
                 handler_pool: CommandHandlerPool):
        self.config = config
        self.telegram_bot = telegram_bot
        self.parser = parser

        self.handlers = {}
        self.is_admin_command = {}
        for h in handler_pool.handlers:
            for card in h.get_cards():
                if card.command_str in self.handlers:
                    raise Exception(f'Handler for {card.command_str} already added: {self.handlers[card.command_str]}')
                self.handlers[card.command_str] = h
                self.is_admin_command[card.command_str] = card.is_admin

    def message(self, message: types.Message) -> bool:
        """
        No need to check for message to be from other bot -- bots in Telegram don't receive messages from other bots
            https://core.telegram.org/bots/faq
        :param message:
        :type message:
        :return:
        :rtype:
        """
        return self.try_handle_command(message)

    def try_handle_command(self, message) -> bool:
        """
        :param message:
        :return: is message a command
        """
        chat_id = message.chat.id

        if message.chat.type == 'channel':
            logger.debug('Message ignored by command handler as it was sent in the channel')
            return False

        if message.content_type != 'text':
            logger.debug('Message ignored by command handler as it is not a text message')
            return False

        text: str = message.text
        if text is None:
            logger.debug('Message ignored by command handler as it does not have text')
            return False

        is_command, command_str = self.parser.is_command(text, return_first_word=True)
        if not is_command:
            logger.debug('Message ignored by command handler as it does not have text')
            return False

        # Hide password if it's in the command
        if self.parser.contains_secret(text):
            try:
                text_w_hide = self.parser.hide_secret(text)
                self.telegram_bot.delete_message(chat_id, message.message_id)
                self.telegram_bot.send_text(chat_id, text_w_hide)
            except ApiTelegramException as ex:
                logger.info(f'Got an exception when trying to hide password, most likely bot does not have admin '
                            f'rights: {ex}')
                self.telegram_bot.send_text(chat_id, 'Please provide admin rights for bot to be able to hide '
                                                     'password in the chat log')

        args = self.parser.parse_command(text)

        if self.parser.error_message is not None:
            logger.debug(f'command handler parser error "{self.parser.error_message}"')
            # Reply in private chats only as in groups multiple bots may be present
            if message.chat.type == 'private':
                response = config_utils.get_response_command_parser_error(config=self.config, command=command_str)
                self.telegram_bot.send_text(chat_id, response)
            return True

        logger.info(f'Bot command: "{args.command}"')

        is_password_correct = self.check_password(args=args, chat_id=chat_id) or \
            self.check_password(args=args, chat_id=message.from_user.id)
        if (message.chat.type != 'private') and (not is_password_correct):
            logger.info(f'Ignoring command as it is from non-admin in non-private chat')
            return True

        has_no_handler = command_str not in self.handlers
        admin_fail_auth = self.is_admin_command[command_str] and (not is_password_correct)
        if has_no_handler or admin_fail_auth:
            response = config_utils.get_response_unknown_command(config=self.config, command=command_str)
            self.telegram_bot.send_text(chat_id=chat_id, text=response)
            return True

        context = CommandContext(telegram_bot=self.telegram_bot,
                                 sender_message=message,
                                 config=self.config,
                                 parser=self.parser,
                                 args=args)
        try:
            self.handlers[command_str].handle(context)
        except CommandMissingArgError as ex:
            context.reply(str(ex), log_level=logging.INFO)
        except Exception as ex:
            logger.exception(ex)
            if is_password_correct:
                context.reply(str(ex), log_level=None)
        return True

    def check_password(self, args, chat_id):
        provided_password = args.password
        admin_password = config_utils.try_get_admin_password(self.config)
        if provided_password is None:
            rem_password = config_utils.try_get_remembered_password(config=self.config, chat_id=chat_id)
            if rem_password is not None:
                if rem_password == admin_password:
                    logger.debug(f'Used remembered password')
                    provided_password = rem_password
                else:
                    logger.debug(f'Remembered password is outdated, forgetting')
                    config_utils.delete_remembered_password(config=self.config, chat_id=chat_id)
        result = (provided_password == admin_password)
        return result



import abc
from typing import Iterable

from tengi.command.card import CommandCard
from tengi.command.command_context import CommandContext


class CommandHandler:
    @abc.abstractmethod
    def get_cards(self) -> Iterable[CommandCard]:
        pass

    @abc.abstractmethod
    def handle(self,
               context: CommandContext):
        pass



from typing import Callable, Optional
import logging

from telebot.types import Message
from tengi.telegram.telegram_bot import TelegramBot

LogFunc = Optional[Callable[[str], None]]

logger = logging.getLogger(__file__)


class ReplyContext:
    def reply(self,
              text: str,
              log_level: Optional[int] = logging.DEBUG,
              stacklevel: int = 2) -> None:
        """
        @param text:
        @type text:
        @param log_level:
        @type log_level:
        @param stacklevel:
        @type stacklevel: stacklevel = 1 is the current function, stacklevel = 2 is the outside calling function
        @return:
        @rtype:
        """
        if log_level is not None:
            # Log using the module name of the previous module
            logger.log(level=log_level, msg=text, stacklevel=stacklevel)


class ReplyContextTelegram(ReplyContext):
    def __init__(self,
                 telegram_bot: TelegramBot,
                 sender_message: Message):
        self.telegram_bot = telegram_bot
        self.sender_message = sender_message

    @property
    def sender_chat_id(self) -> int:
        return self.sender_message.chat.id

    def reply(self,
              text: str,
              log_level: Optional[int] = logging.DEBUG,
              stacklevel: int = 3) -> None:
        """

        @param text:
        @type text:
        @param log_level:
        @type log_level:
        @param stacklevel:
        @type stacklevel: stacklevel = 3 means outside function taking into account call to super().reply
        @return:
        @rtype:
        """
        super().reply(text=text, log_level=log_level, stacklevel=stacklevel)

        if (self.sender_message is None) or (self.sender_message.reply_to_message is None):
            reply_to_message_id = None
        else:
            reply_to_message_id = self.sender_message.reply_to_message.message_id

        self.telegram_bot.send_text(chat_id=self.sender_chat_id,
                                    reply_to_message_id=reply_to_message_id,
                                    text=text)


class ReplyContextLogOnly(ReplyContext):
    pass



from telebot import types
from typing import List
from datetime import datetime
import logging

from tengi.telegram.telegram_bot import TelegramBot

logger = logging.getLogger(__file__)


class TelegramCursor:
    def __init__(self, bot: TelegramBot, look_back_days: float, long_polling_timeout: float = 20):
        self.bot = bot
        self.look_back_days = look_back_days
        self.long_polling_timeout = long_polling_timeout
        self.last_bot_update_id = None

    def look_back(self, allowed_updates):
        updates = self.bot.get_updates(long_polling_timeout=0,
                                       limit=100,
                                       allowed_updates=allowed_updates)

        # Sort updates from newest to oldest to use latest message in the chat only
        updates = sorted(updates,
                         key=lambda upd: upd.update_id,
                         reverse=True)
        now = datetime.utcnow()
        look_back_seconds = self.look_back_days * 24 * 60 * 60
        look_back_updates = []
        cached_chat_ids = set()
        for u in updates:
            if u.message is not None:  # Ignore messages that are outside the look back window
                elapsed_seconds = (now - datetime.utcfromtimestamp(u.message.date)).total_seconds()
                if elapsed_seconds > look_back_seconds:
                    continue
                # Cache only the last message from the chat
                chat_id = u.message.chat.id
                if chat_id in cached_chat_ids:
                    continue
                cached_chat_ids.add(chat_id)
            look_back_updates.append(u)

        # Sort updates from oldest to newest to handle in natural order
        look_back_updates = sorted(look_back_updates,
                                   key=lambda upd: upd.update_id)
        return look_back_updates

    def get_new_updates(self, allowed_updates) -> List[types.Update]:
        look_back_updates = []
        if self.last_bot_update_id is None:
            look_back_updates = self.look_back(allowed_updates=allowed_updates)
            if look_back_updates:
                last_update = max(look_back_updates, key=lambda upd: upd.update_id)
                self.last_bot_update_id = last_update.update_id
            else:
                self.last_bot_update_id = -1

        long_polling_timeout = self.long_polling_timeout if (not look_back_updates) else 0
        updates: List[types.Update] = self.bot.get_updates(offset=self.last_bot_update_id + 1,
                                                           long_polling_timeout=long_polling_timeout,
                                                           allowed_updates=allowed_updates)
        if look_back_updates:
            updates = look_back_updates + updates

        if updates:
            last_update = max(updates, key=lambda upd: upd.update_id)
            self.last_bot_update_id = last_update.update_id

        return updates



from typing import Optional
import time


class AbuseDetector:
    def __init__(self, period_seconds: float, abuse_threshold: int):
        self.period_seconds = period_seconds
        self.abuse_threshold = abuse_threshold
        self.history = {}

    def check_abuse(self, sender) -> Optional[float]:
        s_hist = self._ensure_sender_history(sender)
        cur_time = time.time()
        s_hist = self.select_relevant_history(s_hist=s_hist, cur_time=cur_time)

        result = None
        if len(s_hist) >= self.abuse_threshold:
            oldest_elapsed = cur_time - min(s_hist)
            time_till_oldest_pop = self.period_seconds - oldest_elapsed
            result = time_till_oldest_pop
        else:
            s_hist.append(cur_time)

        self.history[sender] = s_hist
        return result

    def select_relevant_history(self, s_hist, cur_time):
        result = [x for x in s_hist if (cur_time - x <= self.period_seconds) and (cur_time >= x)]
        return result

    def _ensure_sender_history(self, sender):
        if sender not in self.history:
            self.history[sender] = []
        return self.history[sender]



import telebot
from telebot import types
from telebot.apihelper import ApiTelegramException
from typing import Collection, List
import logging

from tengi import event
from tengi.telegram import telegram_error, telegram_bot_utils

logger = logging.getLogger(__file__)

EV_TEL_SENT_MESSAGE = 'telegram.send_message'
EV_TEL_RECEIVED_MESSAGES = 'telegram.received_messages'


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


class TelegramBot:
    def __init__(self, token: str):
        self.bot = telebot.TeleBot(token, parse_mode='HTML')
        self.last_bot_update_id = None

    def send_text(self,
                  chat_id,
                  text,
                  reply_to_message_id=None,
                  buttons: Collection[str] = None,
                  buttons_data: Collection[str] = None,
                  buttons_columns=2):
        reply_markup = self._build_reply_markup(buttons=buttons,
                                                buttons_data=buttons_data,
                                                buttons_columns=buttons_columns)

        sent_message = self.bot.send_message(chat_id=chat_id,
                                             text=text,
                                             reply_to_message_id=reply_to_message_id,
                                             reply_markup=reply_markup)
        event.emitter.emit(EV_TEL_SENT_MESSAGE, sent_message)

    def resend_message(self,
                       chat_id,
                       src_message: types.Message,
                       buttons: Collection[str] = None,
                       buttons_data: Collection[str] = None,
                       buttons_columns=2):
        reply_markup = self._build_reply_markup(buttons=buttons,
                                                buttons_data=buttons_data,
                                                buttons_columns=buttons_columns)
        reply_to_message_id = None if (src_message.reply_to_message is None) \
            else src_message.reply_to_message.message_id

        params_dict = dict(chat_id=chat_id,
                           reply_markup=reply_markup,
                           reply_to_message_id=reply_to_message_id)
        if src_message.text is not None:
            params_dict['text'] = telegram_bot_utils.message_html_text_fixed(src_message)
        elif src_message.caption is not None:
            params_dict['caption'] = telegram_bot_utils.message_html_caption_fixed(src_message)

        if src_message.content_type == 'text':
            send_func = self.bot.send_message
        elif src_message.content_type == 'photo':
            photos = src_message.json['photo']
            params_dict['photo'] = photos[0]['file_id']
            send_func = self.bot.send_photo
        elif src_message.content_type == 'video':
            params_dict['data'] = src_message.video.file_id
            send_func = self.bot.send_video
        elif src_message.content_type == 'audio':
            params_dict['audio'] = src_message.audio.file_id
            send_func = self.bot.send_audio
        elif src_message.content_type == 'document':
            params_dict['data'] = src_message.document.file_id
            send_func = self.bot.send_document
        elif src_message.content_type == 'sticker':
            params_dict['data'] = src_message.sticker.file_id
            send_func = self.bot.send_sticker
        elif src_message.content_type == 'video_note':
            params_dict['data'] = src_message.video_note.file_id
            send_func = self.bot.send_video_note
        elif src_message.content_type == 'voice':
            params_dict['voice'] = src_message.voice.file_id
            send_func = self.bot.send_voice
        elif src_message.content_type == 'location':
            params_dict['location'] = src_message.location
            send_func = self.bot.send_location
        elif src_message.content_type == 'contact':
            params_dict['phone_number'] = src_message.contact.phone_number
            params_dict['first_name'] = src_message.contact.first_name
            params_dict['last_name'] = src_message.contact.last_name
            params_dict['vcard'] = src_message.contact.vcard
            send_func = self.bot.send_contact
        else:
            raise ValueError(f'Resending of {src_message.content_type} content not supported')

        assert send_func is not None
        sent_message = send_func(**params_dict)
        event.emitter.emit(EV_TEL_SENT_MESSAGE, sent_message)

    def edit_message_text(self,
                          text,
                          content_type,
                          chat_id,
                          message_id):
        if content_type == 'text':
            self.bot.edit_message_text(text=text,
                                       chat_id=chat_id,
                                       message_id=message_id)
        else:
            self.bot.edit_message_caption(chat_id=chat_id,
                                          message_id=message_id,
                                          caption=text)

    def forward_message(self, to_chat_id, from_chat_id, message_id):
        sent_message = self.bot.forward_message(to_chat_id, from_chat_id, message_id)
        event.emitter.emit(EV_TEL_SENT_MESSAGE, sent_message)

    def delete_message(self, chat_id, message_id):
        self.bot.delete_message(chat_id, message_id)

    def answer_callback_query(self, callback_query_id, text=None):
        self.bot.answer_callback_query(callback_query_id=callback_query_id, text=text)

    def get_updates(self,
                    offset=None,
                    limit=None,
                    long_polling_timeout=20,
                    try_delete_webhook=True,
                    allowed_updates=None):
        """
        :param offset:
        :param limit:
        :param long_polling_timeout:
        :param try_delete_webhook: fixes an issue with a parallel bot setup a webhook which is incompatible approach to
                   get_updates
        :param allowed_updates:
        :return:
        """
        try:
            updates: List[types.Update] = self.bot.get_updates(offset=offset,
                                                               limit=limit,
                                                               long_polling_timeout=long_polling_timeout,
                                                               allowed_updates=allowed_updates)
        except ApiTelegramException as ex:
            if try_delete_webhook and (ex.error_code == telegram_error.CONFLICT):
                logger.info(f'Got Telegram exception on get updates, will try to remove webhook & repeat: {ex}')
                self.bot.delete_webhook(drop_pending_updates=False)
                updates = self.get_updates(offset=offset,
                                           limit=limit,
                                           long_polling_timeout=long_polling_timeout,
                                           try_delete_webhook=False)
            else:
                raise ex

        received_messages = [u.message for u in updates if u.message is not None]
        if received_messages:
            event.emitter.emit(EV_TEL_RECEIVED_MESSAGES, received_messages)

        return updates

    @staticmethod
    def _build_reply_markup(buttons: Collection[str],
                            buttons_data: Collection[str],
                            buttons_columns: int):
        reply_markup = None
        if buttons is not None:
            if buttons_data is None:
                buttons_data = buttons
            assert len(buttons) == len(buttons_data), \
                f'Buttons & buttons data size mismatch: {len(buttons)} != {len(buttons_data)}'

            reply_markup = types.InlineKeyboardMarkup()

            buttons_obj = [types.InlineKeyboardButton(text=text, callback_data=data)
                           for text, data in zip(buttons, buttons_data)]
            buttons_rows = list(chunks(buttons_obj, n=buttons_columns))

            for row in buttons_rows:
                reply_markup.add(*row)
        return reply_markup



from typing import Iterable
from telebot.types import Message
from tengi import event
from tengi.telegram.telegram_bot import EV_TEL_RECEIVED_MESSAGES
from tengi.state.preserver import *


class ChatIdPreserver(Preserver):
    def __init__(self, state_file_path: Path):
        super().__init__(state_file_path)

        event.emitter.on(EV_TEL_RECEIVED_MESSAGES, self._on_receive_messages)

    def _ensure_private_chats_list(self) -> list:
        if 'private_chats' not in self.state:
            self.state['private_chats'] = []

        return self.state['private_chats']

    def _update_private_chats_list(self, value: list):
        self.state['private_chats'] = value

    def _add_private_chat(self, chat_id):
        private_chats = self._ensure_private_chats_list()
        if chat_id not in private_chats:
            private_chats.append(chat_id)
            self._update_private_chats_list(private_chats)

    def _on_receive_messages(self, messages: Iterable[Message]):
        for m in messages:
            if m.chat.type == 'private':
                self._add_private_chat(m.chat.id)



from setuptools import setup, find_packages

setup(
    name='tengi',
    version='dev',
    url='',
    author='dev',
    author_email='',
    description='',
    packages=find_packages(),    
    install_requires=[],
)



import asyncio
from telethon.sync import TelegramClient
from telethon.tl.types import Message
from typing import cast, List

from tengi.telegram import telegram_bot_utils


class TelegramApi:
    def __init__(self, api_session_name, api_id, api_hash):
        """
        :param api_session_name: arbitrary string to distinguish between your sessions
        :param api_id: obtain at https://my.telegram.org/auth
        :param api_hash: obtain at https://my.telegram.org/auth
        """
        self.api_session_name = api_session_name
        self.api_id = api_id
        self.api_hash = api_hash
        asyncio.run(self._setup_session())

    def _get_api_client(self):
        return TelegramClient(session=self.api_session_name,
                              api_id=self.api_id,
                              api_hash=self.api_hash)

    async def _setup_session(self):
        """
        Creates a session object and caches it on disk, so next requests will reuse it and don't ask for phone number
        """
        api_client = self._get_api_client()
        async with api_client:
            pass

    def get_chat_message(self, chat_id, message_id) -> Message:
        return asyncio.run(self.get_chat_message_async(chat_id=chat_id,
                                                       message_id=message_id))

    async def get_chat_message_async(self, chat_id, message_id) -> Message:
        if type(chat_id) == str:
            chat_id = telegram_bot_utils.to_int_chat_id_if_possible(chat_id)

        api_client = self._get_api_client()
        async with api_client:
            msg = await api_client.get_messages(chat_id, ids=message_id)

        if msg is None:
            raise Exception(f'In chat {chat_id} there is no message {message_id}')

        # api_client.get_messages returns Message object in case of 1 message requested though TotalList declared
        msg = cast(Message, msg)
        return msg

    def get_chat_messages_backward(self, chat_id, message_id, n_messages) -> List[Message]:
        return asyncio.run(self.get_chat_messages_backward_async(chat_id=chat_id,
                                                                 message_id=message_id,
                                                                 n_messages=n_messages))

    async def get_chat_messages_backward_async(self, chat_id, message_id, n_messages) -> List[Message]:
        if type(chat_id) == str:
            chat_id = telegram_bot_utils.to_int_chat_id_if_possible(chat_id)

        api_client = self._get_api_client()
        async with api_client:
            arr_messages = []
            gen = api_client.iter_messages(chat_id, max_id=message_id+1, limit=n_messages)
            async for msg in gen:
                arr_messages.append(msg)
        return arr_messages

    def get_chat_last_messages(self, chat_id, max_messages, max_excluded_id=0) -> List[Message]:
        return asyncio.run(self.get_chat_last_messages_async(chat_id=chat_id,
                                                             max_messages=max_messages,
                                                             max_excluded_id=max_excluded_id))

    async def get_chat_last_messages_async(self, chat_id, max_messages, max_excluded_id=0) -> List[Message]:
        api_client = self._get_api_client()
        async with api_client:
            arr_messages = []
            async for msg in api_client.iter_messages(chat_id, limit=max_messages, min_id=max_excluded_id):
                arr_messages.append(msg)
        return arr_messages



from tengi import CommandParam

params = [
    CommandParam(name='--password',
                 help_str='Password to access the bot admin commands',
                 param_type=str),
    CommandParam(name='--name',
                 help_str='Name of a variable to read',
                 param_type=str),
    CommandParam(name='--value',
                 help_str='Value to set',
                 param_type=str),
    CommandParam(name='--chat_id',
                 help_str='Chat id to perform operation on',
                 param_type=str),
]



import json
from json import JSONDecodeError
from telebot import types
from typing import Union, Iterable, List, Optional


def encode_button_data(handler: str,
                       case_id: str,
                       response: str) -> str:
    data_dict = dict(h=handler,
                     c=case_id,
                     r=response)
    data_str = json.dumps(data_dict, separators=(',', ':'))
    data_bytes = data_str.encode('utf-8')
    # Keyboard button data should be <=64 bytes https://core.telegram.org/bots/api#inlinekeyboardbutton
    assert len(data_bytes) <= 64, f'data is too long, {len(data_bytes)} > 64'
    return data_str


def is_button_data_encoded(data: str) -> bool:
    try:
        data_dict = json.loads(data)
        result = all([key in data_dict for key in ['h', 'c', 'r']])
    except JSONDecodeError:
        result = False
    return result


def decode_button_data(data: str):
    data_dict = json.loads(data)
    return data_dict['h'], data_dict['c'], data_dict['r']


def encode_case_id(chat_id, message_id) -> str:
    return f'{chat_id}_{message_id}'


def decode_case_id(case_id: str):
    parts = case_id.split('_')
    assert len(parts) == 2
    chat_id, message_id = int(parts[0]), int(parts[1])
    return chat_id, message_id


def flat_trim_message(message: types.Message, max_len: int, terminator='…'):
    text = message.text
    if text is None:
        text = message.caption

    if text is not None:
        # Make flat
        text = ' '.join(text.split())
        if len(text) > max_len:
            trimmed_len = max_len - len(terminator)
            text = text[:trimmed_len] + terminator
    return text


def get_file_id(message: types.Message):
    file_id = None
    if message.content_type == 'photo':
        photos = message.json['photo']
        file_id = photos[0]['file_id']
    elif message.content_type == 'video':
        file_id = message.video.file_id
    elif message.content_type == 'audio':
        file_id = message.audio.file_id
    elif message.content_type == 'document':
        file_id = message.document.file_id
    elif message.content_type == 'sticker':
        file_id = message.sticker.file_id
    elif message.content_type == 'video_note':
        file_id = message.video_note.file_id
    elif message.content_type == 'voice':
        file_id = message.voice.file_id
    elif message.content_type == 'animation':
        file_id = message.animation.file_id
    return file_id


def message_contains_link(message: types.Message, link_entities=None):
    if link_entities is None:
        link_entities = ['text_link', 'url', 'phone_number', 'mention', 'email']

    return message_contains_entity(message, entity_types=link_entities)


def message_contains_entity(message: types.Message, entity_types: Iterable[str]) -> bool:
    message_entities: List[dict] = list()
    if 'entities' in message.json:
        message_entities.extend(message.json['entities'])
    if 'caption_entities' in message.json:
        message_entities.extend(message.json['caption_entities'])

    has_one_of_entities = any([m['type'] in entity_types for m in message_entities])
    return has_one_of_entities


def is_private_message(message: types.Message):
    if (message.chat is None) or (message.chat.type != 'private'):
        return False
    return True


def is_group_message(message: types.Message):
    if (message.chat is None) or (message.chat.type not in ['group', 'supergroup']):
        return False
    return True


def is_int_chat_id(str_chat_id: str):
    try:
        int(str_chat_id)
        result = True
    except ValueError:
        result = False
    return result


def is_proper_chat_id(str_chat_id: str):
    result = is_int_chat_id(str_chat_id) or \
             (str_chat_id and (str_chat_id[0] == '@'))
    return result


def to_int_chat_id_if_possible(str_chat_id: str) -> Union[str, int]:
    result = str_chat_id
    if is_int_chat_id(str_chat_id):
        result = int(str_chat_id)
    return result


def get_short_chat_id(chat_id: int):
    result = chat_id
    str_chat_id = str(chat_id)
    prefix = '-100'
    if str_chat_id.startswith(prefix):
        short_str_chat_id = str_chat_id[len(prefix):]
        result = int(short_str_chat_id)
    return result


def escape_html_tags(text: str) -> str:
    """
    Current escaping is a hotfix. The issue with replacement is that with a proper symbol escaping the message entities
    (offset and length) should also be updated.
    The hotfix idea is  to remove special symbols and keep message length the same.
    Proper escaping should do:
    & -> &amp;
    < -> &lt;
    > -> &gt;
    """
    text = text \
        .replace('&', '_') \
        .replace('<', '_') \
        .replace('>', '_')
    return text


def message_html_text_fixed(msg: types.Message) -> Optional[str]:
    if msg.text is None:
        return None
    msg_copy = types.Message.de_json(msg.json)
    msg_copy.text = escape_html_tags(msg_copy.text)
    return msg_copy.html_text


def message_html_caption_fixed(msg: types.Message) -> Optional[str]:
    if msg.caption is None:
        return None
    msg_copy = types.Message.de_json(msg.json)
    msg_copy.caption = escape_html_tags(msg_copy.caption)
    return msg_copy.html_caption



import logging
from typing import Optional

from tengi.setup.config import Config

logger = logging.getLogger(__file__)


def get_protected_vars(config: Config) -> list:
    result = config['protected_vars'] if ('protected_vars' in config) else []
    return result


def is_variable_protected(config: Config, variable: str) -> bool:
    protected_vars = get_protected_vars(config)
    return variable in protected_vars


def get_remembered_passwords(config: Config) -> dict:
    result = config['remembered_passwords'] if ('remembered_passwords' in config) else dict()
    return result


def update_remembered_passwords(config: Config, remembered_passwords: dict):
    config['remembered_passwords'] = remembered_passwords


def delete_remembered_password(config: Config, chat_id):
    assert has_remembered_password(config=config, chat_id=chat_id)
    remembered_passwords = get_remembered_passwords(config)
    str_chat_id = str(chat_id)
    del remembered_passwords[str_chat_id]
    update_remembered_passwords(config=config, remembered_passwords=remembered_passwords)


def has_remembered_password(config: Config, chat_id) -> bool:
    remembered_passwords = get_remembered_passwords(config)
    str_chat_id = str(chat_id)
    return str_chat_id in remembered_passwords


def remember_password(config: Config, chat_id, password: str):
    remembered_passwords = get_remembered_passwords(config)
    str_chat_id = str(chat_id)
    remembered_passwords[str_chat_id] = password
    update_remembered_passwords(config=config, remembered_passwords=remembered_passwords)


def try_get_remembered_password(config: Config, chat_id) -> Optional[str]:
    arr_rem_passwords = get_remembered_passwords(config)
    chat_id_str = str(chat_id)
    rem_password = arr_rem_passwords.get(chat_id_str, None)
    return rem_password


def try_get_admin_password(config: Config):
    return config.try_get_warny('admin_password', operation_name='control access to admin commands')


def get_response_command_parser_error(config: Config, command: str):
    format_str = config['response_command_parser_error'] if ('response_command_parser_error' in config) else \
        '{command} error'
    result = format_str.format(command=command)
    return result


def get_response_unknown_command(config: Config, command: str):
    format_str = config['response_unknown_command'] if ('response_unknown_command' in config) else '{command} unknown'
    result = format_str.format(command=command)
    return result


def try_get_response_start(config: Config) -> Optional[str]:
    return config.try_get_warny('response_start', operation_name='respond to /start')


def try_get_response_help(config: Config) -> Optional[str]:
    return config.try_get_warny('response_help', operation_name='respond to /help')



import logging
import json

from tengi.command.command_handler import *
from tengi.setup import config_utils

logger = logging.getLogger(__file__)


class CommandHandlerConfig(CommandHandler):
    def get_cards(self) -> Iterable[CommandCard]:
        return [CommandCard(command_str='/get_config',
                            description='Admin. Reads config variable',
                            is_admin=True),
                CommandCard(command_str='/set_config',
                            description='Admin. Sets config variable',
                            is_admin=True)]

    def handle(self, context: CommandContext):
        var_name = context.get_mandatory_arg('name')

        if config_utils.is_variable_protected(config=context.config, variable=var_name):
            logger.info(f'Received {context.command} command with protected --name {var_name}')
            context.reply(f'{var_name} protected')
            return

        if var_name not in context.config:
            logger.info(f'Received {context.command} command with unknown --name {var_name}')
            context.reply(f'{var_name} not in config')
            return

        var_old_value = context.config[var_name]

        if context.command == '/get_config':
            context.reply(str(var_old_value), log_level=logging.INFO)
        elif context.command == '/set_config':
            var_new_value = context.get_mandatory_arg('value')

            if type(var_old_value) != str:
                var_new_value = json.loads(var_new_value)

            if type(var_old_value) != type(var_new_value):
                context.reply(f'Value types mismatch: {var_old_value}, {var_new_value}',
                              log_level=logging.INFO)
                return

            context.config[var_name] = var_new_value
            context.reply(f'Value {var_name} set successfully', log_level=logging.INFO)
            return
        else:
            raise Exception(f'Unhandled command {context.command} (programmer mistake in if-cascade)')




import signal
import sys
import time
import logging
from typing import Iterable, Callable

logger = logging.getLogger(__file__)


class App:
    def __init__(self,
                 update_funcs: Iterable[Callable],
                 update_seconds: float,
                 restart_seconds: float):
        self.update_funcs = update_funcs
        self.update_seconds = update_seconds
        self.restart_seconds = restart_seconds

    def run(self):
        # KeyboardInterrupt catching doesn't work as telebot catches it internally and sends as a different error
        def sigint_handler(_sig, _frame):
            logger.info('Ctrl+C pressed')
            logger.info('Bot finished')
            sys.exit(0)

        signal.signal(signal.SIGINT, sigint_handler)

        logger.info('App started')
        should_interrupt = False
        while True:
            try:
                while True:
                    upd_begin = time.time()

                    for u_func in self.update_funcs:
                        u_func()

                    upd_end = time.time()
                    upd_remaining = self.update_seconds - (upd_end - upd_begin)

                    should_interrupt = self.should_interrupt()
                    if should_interrupt:
                        logger.info(f'Got should_interrupt')
                        break

                    if upd_remaining > 0:
                        time.sleep(upd_remaining)
            except Exception as e:
                logger.exception(e)

            if should_interrupt:
                break

            logger.info(f'Restarting in {self.restart_seconds:,} seconds')
            time.sleep(self.restart_seconds)

    @classmethod
    def should_interrupt(cls):
        """
        An ability to interrupt infinite loop, used for testing
        :return:
        """
        return False



from jsonstore import JsonStore
import json
from pathlib import Path
import logging

logger = logging.getLogger(__file__)


class Config(JsonStore):
    def __init__(self, config_path: Path, example_path: Path):
        if not example_path.is_file():
            raise FileNotFoundError(f'Cannot find config example {example_path}')
        if not config_path.is_file():
            raise FileNotFoundError(f'Please create {config_path}, use {example_path} as an example')
        super().__init__(str(config_path))

        self.assert_keys_present(example_path)

    def assert_keys_present(self, example_path):
        with open(example_path, 'r') as f:
            example: dict = json.load(f)
        missing_keys = []
        for k in example.keys():
            if k not in self:
                missing_keys.append(k)

        assert len(missing_keys) == 0, f'Add values {missing_keys} to the config {self._path}. ' \
                                       f'Reference: {example_path}'

    @property
    def path_str(self) -> str:
        return self._path

    def try_get_warny(self, key: str, operation_name: str):
        if key in self:
            return self[key]
        else:
            logger.warning(f'Value "{key}" not present in the config. Add it into {self.path_str} to '
                           f'"{operation_name}"')



from jsonstore import JsonStore
from pathlib import Path


class Preserver:
    def __init__(self, state_file_path: Path, indent=2, auto_commit=True):
        state_file_path.parent.mkdir(exist_ok=True, parents=True)
        self.state = JsonStore(str(state_file_path),
                               indent=indent,
                               auto_commit=auto_commit)

    def _get_state_data(self) -> dict:
        # noinspection PyProtectedMember
        return self.state._data

    def _save_state(self):
        # noinspection PyProtectedMember
        self.state._save()
        
    def as_dict(self) -> dict:
        return self.state.__dict__['_data'].copy()



import argparse
import shlex
import re
from typing import Iterable
import itertools

from tengi.command.param import CommandParam
from tengi.command.handler_pool import CommandHandlerPool


class CommandParser(argparse.ArgumentParser):
    secret_pattern_str = r'(?i)([^\w_]+(password|bot_token)\s+)(\S+)'
    secret_pattern = re.compile(secret_pattern_str)

    def __init__(self, handler_pool: CommandHandlerPool, params: Iterable[CommandParam]):
        super().__init__(add_help=False, formatter_class=argparse.RawTextHelpFormatter)

        self.error_message = None

        cards = itertools.chain(*[h.get_cards() for h in handler_pool.handlers])
        arr_command_str = [x.command_str for x in cards]
        self.add_argument('command',
                          choices=arr_command_str,
                          type=str.lower)

        for p in params:
            p.add_to_parser(self)

        self.usage = f'[command] [parameters]'

    def validate_and_fix_params(self, params):
        pass  # Empty for now

    @classmethod
    def unify_whitespace(cls, s):
        # Ws can be obtained using the following code
        #       all_unicode = ''.join(chr(c) for c in range(sys.maxunicode+1))
        #       ws = ''.join(re.findall(r'\s', all_unicode))
        ws = '\t\n\x0b\x0c\r\x1c\x1d\x1e\x1f \x85\xa0\u1680\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008' \
             '\u2009\u200a\u2028\u2029\u202f\u205f\u3000'
        unified_ws = ' '
        s2 = s.translate(s.maketrans(ws, unified_ws * len(ws)))
        return s2

    def parse_command(self, command):
        try:
            # We unify whitespace as shlex can't parse Unicode whitespaces properly
            command = self.unify_whitespace(command)
            # Fix back Telegram auto replace
            command = command.replace('—', '--')
            args = shlex.split(command)

            return self.parse_args(args)
        except Exception as e:
            self.error_message = str(e)

    def parse_args(self, *args, **kwargs):
        self.error_message = None
        try:
            params = super().parse_args(*args, **kwargs)
            # Make checks for command mandatory params to be set
            self.validate_and_fix_params(params)
            return params
        except Exception as e:
            self.error_message = str(e)

    def is_command(self, text, return_first_word=False):
        if text is None:
            return False
        text_cropped = self.unify_whitespace(text)
        text_cropped = text_cropped.strip()
        first_word = text_cropped.split(' ')[0]
        is_command = first_word[0] == '/'
        if return_first_word:
            return is_command, first_word
        else:
            return is_command

    def contains_secret(self, text):
        if text is None:
            return False
        result = self.secret_pattern.search(text) is not None
        return result

    def hide_secret(self, text):
        # \1 represents text before password(or other secret), \2 (the replaced text) -- the password itself
        text_w_hide = self.secret_pattern.sub(r'\1[HIDDEN]', text)
        return text_w_hide

    def error(self, message):
        """
        Overwrite the default parser behaviour in case of error - throws an exception instead of making exit
        """
        message = f'{message}\n*use "help" command to get help*'
        raise Exception(message)

    def format_help(self):
        help_str = super().format_help()
        # default "optional arguments" name isn't suitable for us as some of these arguments are mandatory for some
        #   commands
        help_str = help_str.replace("optional arguments", "parameters")
        # in our case "positional arguments" are used to set the command
        help_str = help_str.replace("positional arguments", "command")
        return help_str



import logging

from tengi.command.command_handler import *
from tengi.setup import config_utils

logger = logging.getLogger(__file__)


class CommandHandlerEssentials(CommandHandler):
    def get_cards(self) -> Iterable[CommandCard]:
        return [CommandCard(command_str='/start',
                            description='Start working with a bot',
                            is_admin=False),
                CommandCard(command_str='/help',
                            description='User help message',
                            is_admin=False),
                CommandCard(command_str='/help2',
                            description='Admin. Print help for the commands',
                            is_admin=True),
                CommandCard(command_str='/ping',
                            description='Checks if the bot is working',
                            is_admin=False),
                CommandCard(command_str='/chat_id',
                            description='Get the current chat id',
                            is_admin=True),
                ]

    def handle(self, context: CommandContext):
        if context.command == '/start':
            response_start = config_utils.try_get_response_start(context.config)
            if response_start is not None:
                context.reply(response_start)

        elif context.command == '/help':
            response_help = config_utils.try_get_response_help(context.config)
            if response_help is not None:
                context.reply(response_help)
        elif context.command == '/help2':
            commands_help_message = '<pre>' + context.parser.format_help() + '</pre>'
            context.reply(commands_help_message)
        elif context.command == '/ping':
            context.reply('pong')
        elif context.command == '/chat_id':
            context.reply(str(context.sender_chat_id))
        else:
            raise ValueError(f'Unhandled command: {context.command}')



import logging
from typing import Iterable, Callable, Any, List
import inspect
from telebot.apihelper import ApiTelegramException

from tengi import TelegramCursor
from tengi.telegram.inbox_handler import *
from tengi.telegram import  telegram_error


logger = logging.getLogger(__file__)


def is_overwritten(method):
    #  __qualname__ examples: 'TelegramInboxHandler.edited_message', 'CommandHub.message'
    return TelegramInboxHandler.__name__ not in method.__qualname__


class TelegramInboxHub:
    def __init__(self, telegram_cursor: TelegramCursor, chain_handlers: Iterable[TelegramInboxHandler]):
        self.telegram_cursor = telegram_cursor

        # getmembers returns array of tuples like
        #   ('callback_query', <function __main__.TelegramInboxHandler.callback_query(self, message:
        #   telebot.types.CallbackQuery) -> Union[bool, NoneType]>)
        update_types = [x[0] for x in inspect.getmembers(TelegramInboxHandler, predicate=inspect.isfunction)]

        # Group handlers in dict of lists by update types
        self.handlers = {}
        for h in chain_handlers:
            for ut in update_types:
                method = getattr(h, ut)
                if is_overwritten(method):
                    if ut not in self.handlers:
                        self.handlers[ut] = []
                    self.handlers[ut].append(h)

    def update(self):
        allowed_updates = list(self.handlers.keys())
        try:
            updates = self.telegram_cursor.get_new_updates(allowed_updates=allowed_updates)
        except ApiTelegramException as ex:
            if ex.error_code == telegram_error.BAD_GATEWAY:
                #  Telegram is temporary unavailable
                logger.info('Telegram is temporary unavailable, further update is skipped')
                return
            else:
                raise ex

        for u in updates:
            try:
                arr_update_type = [ut for ut in allowed_updates if (getattr(u, ut) is not None)]
                assert len(arr_update_type) == 1, f'Update object contains multiple update types: {arr_update_type}'
                update_type = arr_update_type[0]

                def handler_func(handler: TelegramInboxHandler, item_: Any) -> bool:
                    func = getattr(handler, update_type)
                    handled_ = func(item_)
                    return handled_

                # Names of methods in TelegramInboxHandler match names of fields in types.Update
                item = getattr(u, update_type)
                handled = self._chain_handlers(handlers=self.handlers[update_type],
                                               item=item,
                                               handler_func=handler_func)
                if not handled:
                    logger.info(f'Update not handled: {update_type}, {u.update_id}')
            except Exception as ex:
                logger.exception(ex)

    @staticmethod
    def _chain_handlers(handlers: List[TelegramInboxHandler],
                        item: Any,
                        handler_func: Callable[[TelegramInboxHandler, Any], bool]):
        handled = False
        for h in handlers:
            handled = handler_func(h, item)
            if handled:
                break
        return handled



import jsonstore
from functools import partial


def fix_jsonstore_dumps():
    # Fix an issue of jsonstore dumping non-ASCII ugly text, like '\u2321'
    jsonstore.json.dumps = partial(jsonstore.json.dumps, ensure_ascii=False)



import csv
from datetime import datetime
from pathlib import Path
from typing import Optional, Tuple


class CsvLogger:
    def __init__(self, dir_path, file_name_prefix):
        self.dir_path = dir_path
        self.file_name_prefix = file_name_prefix

        self.date = None
        self.csv_file = None
        self.writer = None
        self.columns_set = None

    def write_row(self, row):
        self._ensure_writer(new_col_names=row.keys())
        self.writer.writerow(row)
        self.csv_file.flush()

    def _clean_state(self):
        if self.csv_file is not None:
            self.csv_file.close()
        self.csv_file = None
        self.writer = None
        self.columns_set = None

    def _ensure_writer(self, new_col_names) -> None:
        today = datetime.utcnow().date()
        if self.date != today:  # None or other date
            self._clean_state()

            self.date = today
            file_path, _ = self._find_date_latest_file(self.date)

            # More complex, check format change options also
            if file_path is not None:
                with open(file_path, 'r') as infile:
                    reader = csv.DictReader(infile)
                    old_col_names = reader.fieldnames

                if old_col_names is not None:
                    self.columns_set = set(old_col_names)
                    self.csv_file = open(file_path, 'a')
                    self.writer = csv.DictWriter(self.csv_file, fieldnames=old_col_names, dialect='excel')

        if (self.writer is None) or (not set(new_col_names).issubset(self.columns_set)):
            # Create a new file
            self._clean_state()
            file_path = self._create_date_next_file(today)
            self.columns_set = set(new_col_names)
            file_path.parent.mkdir(exist_ok=True, parents=True)
            self.csv_file = open(file_path, 'w')
            self.writer = csv.DictWriter(self.csv_file, fieldnames=new_col_names, dialect='excel')
            self.writer.writeheader()
            self.csv_file.flush()

    def _find_date_latest_file(self, date) -> Tuple[Optional[Path], Optional[int]]:
        prefix_w_date = f'{self.file_name_prefix}_{date}'
        files = list(self.dir_path.glob(f'{prefix_w_date}*'))
        if len(files) == 0:
            return None, None
        else:
            suffixes = [x.name.replace(prefix_w_date, '') for x in files]
            suffixes = [x.replace('.csv', '') for x in suffixes]
            num_suffixes = [x.replace('_', '') for x in suffixes]
            num_suffixes = [int(x) if x.isdigit() else None for x in num_suffixes]

            num_suffixes_w_value = [x for x in num_suffixes if x is not None]

            if len(num_suffixes_w_value) == 0:
                file_idx = suffixes.index('') if '' in suffixes else None
                if file_idx is not None:
                    return files[file_idx], None
                else:
                    return None, None
            else:
                variant_num = max(num_suffixes_w_value)
                file_idx = num_suffixes.index(variant_num)
                return files[file_idx], variant_num

    def _create_date_next_file(self, date) -> Path:
        cur_latest_file, latest_variant_num = self._find_date_latest_file(date)

        if cur_latest_file is None:
            file_name = f'{self.file_name_prefix}_{date}.csv'
        else:
            variant_num = latest_variant_num + 1 if latest_variant_num is not None else 1
            file_name = f'{self.file_name_prefix}_{date}_{variant_num}.csv'
        return self.dir_path / file_name



from telebot import types


class TelegramInboxHandler:
    """
    Methods of this class should be exactly the same as different options of telebot.types.Update
        see https://core.telegram.org/bots/api#update
    Don't add any other functions. In case you need some middleware logic -- create a subclass of this one.
    If  my_chat_member, chat_member will be needed - they are likely sent as Message, see telebot.types.Message.de_json
    """

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def message(self, message: types.Message) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def edited_message(self, message: types.Message) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def channel_post(self, channel_post: types.Message) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def edited_channel_post(self, edited_channel_post: types.Message) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def inline_query(self, inline_query: types.InlineQuery) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def chosen_inline_result(self, chosen_inline_result: types.ChosenInlineResult) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def callback_query(self, callback_query: types.CallbackQuery) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def shipping_query(self, shipping_query: types.ShippingQuery) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def pre_checkout_query(self, pre_checkout_query: types.PreCheckoutQuery) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def poll(self, poll: types.Poll) -> bool:
        return False

    # noinspection PyUnusedLocal, PyMethodMayBeStatic
    def poll_answer(self, poll_answer: types.PollAnswer) -> bool:
        return False



import time
import logging

from tengi.state.abuse_detector import AbuseDetector

logger = logging.getLogger(__file__)


class AbuseJanitor:
    """
    Periodically cleans abuse detector state from outdated user records (prevent RAM overfill)
    """
    def __init__(self, abuse_detector: AbuseDetector, period_seconds: float):

        self.abuse_detector = abuse_detector
        self.period_seconds = period_seconds
        self.last_run = time.time()

    def update(self):
        cur_time = time.time()
        if (cur_time - self.last_run) >= self.period_seconds:
            logger.info('janitor run')

            for sender, s_hist in list(self.abuse_detector.history.items()):
                s_hist = self.abuse_detector.select_relevant_history(s_hist=s_hist, cur_time=cur_time)
                if not s_hist:
                    del self.abuse_detector.history[sender]

            self.last_run = cur_time



from argparse import Namespace
from typing import Any, TYPE_CHECKING
from tengi.telegram.reply_context import *
from tengi.setup.config import Config
from tengi.command.command_error import CommandMissingArgError

if TYPE_CHECKING:
    from tengi.command.command_parser import CommandParser


class CommandContext(ReplyContextTelegram):
    def __init__(self,
                 telegram_bot: TelegramBot,
                 sender_message: Message,
                 config: Config,
                 parser: 'CommandParser',
                 args: Namespace):
        super().__init__(telegram_bot=telegram_bot, sender_message=sender_message)
        self.config = config
        self.parser = parser
        self.args = args

    @property
    def command(self):
        return self.args.command

    def get_mandatory_arg(self, arg_name: str, cast_func: Optional[Callable] = None) -> Any:
        value = getattr(self.args, arg_name)
        if value is None:
            raise CommandMissingArgError(f'--{arg_name} required')
        if cast_func is not None:
            try:
                value = cast_func(value)
            except ValueError:
                raise CommandMissingArgError(f'--{arg_name} should have type {cast_func}')
        return value

    def get_optional_arg(self, arg_name: str,
                         default: Any,
                         cast_func: Optional[Callable] = None) -> Any:
        value = getattr(self.args, arg_name, default)
        if cast_func is not None:
            try:
                value = cast_func(value)
            except ValueError:
                raise CommandMissingArgError(f'--{arg_name} should have type {cast_func}')
        return value



import hashlib
from typing import Any, Optional
import base64
import logging

from tengi.setup.config import Config

logger = logging.getLogger(__file__)


class Hasher:
    def __init__(self, config: Config):
        self.config = config

    def trimmed(self, v: Any, hash_bytes=None) -> Optional[str]:
        """
        param hash_bytes: how many bytes to hake for the trimmed hash, 10 bytes of data -> 16 symbols string
        """
        if v is None:
            return None

        s = str(v)
        salt = self.config.try_get_warny('hash_salt', operation_name='improve hashed data protection')
        if (salt is not None) and (salt != ''):
            s = s + salt

        if hash_bytes is None:
            hash_bytes = self.config.try_get_warny('hash_bytes', operation_name='reduce hashes length')

        if hash_bytes <= 0:
            logger.warning(f'Hash bytes value should be >0, but not "{hash_bytes}", ignoring it')
            hash_bytes = None

        s_bytes = s.encode('utf-8')
        hash_result = hashlib.sha1(s_bytes)
        bytes_taken = hash_result.digest() if hash_bytes is None else hash_result.digest()[-hash_bytes:]
        result = base64.b32encode(bytes_taken).decode('utf-8').rstrip('=')
        return result





class CommandCard:
    def __init__(self, command_str: str, description: str, is_admin: bool):
        self.command_str = command_str
        self.description = description
        self.is_admin = is_admin



import time
import copy

from tengi.state.preserver import *


class TimedPreserver(Preserver):
    def __init__(self, state_file_path: Path, save_period: float, indent=2):
        super().__init__(state_file_path=state_file_path,
                         indent=indent,
                         auto_commit=False)
        self.save_period = save_period
        self.saved_time = time.time()
        self.saved_dict = copy.deepcopy(self._get_state_data())

    def update(self):
        cur_time = time.time()
        if (self.saved_time is None) or \
                (cur_time - self.saved_time >= self.save_period):

            cur_dict = self._get_state_data()
            if self.saved_dict != cur_dict:
                self._save_state()
                self.saved_dict = copy.deepcopy(cur_dict)

            self.saved_time = cur_time




# Taken from: https://core.telegram.org/api/errors#400-bad-request
#   some (like CONFLICT) are googled
BAD_REQUEST = 400
UNAUTHORIZED = 401
FORBIDDEN = 403  # User blocked the bot
NOT_FOUND = 404
NOT_ACCEPTABLE = 406
CONFLICT = 409
FLOOD = 420
TOO_MANY_REQUESTS = 429
INTERNAL = 500
BAD_GATEWAY = 502  # Telegram is temporary unavailable



from typing import Any
import argparse


class CommandParam:
    def __init__(self, name: str, help_str: str, param_type: Any, nargs=None):
        assert name.startswith('--')

        self.name = name
        self.help_str = help_str
        self.param_type = param_type
        self.nargs = nargs

    def add_to_parser(self, parser: argparse.ArgumentParser):
        parser.add_argument(self.name,
                            type=self.param_type,
                            nargs=self.nargs,
                            metavar='',
                            help=self.help_str)
